from fastgenomics.defaults import DEFAULT_APP_DIR, DEFAULT_DATA_ROOT, INPUT_FILE_MAPPING
from fastgenomics.process import FGProcess
import pytest
from pathlib import Path


def test_values():
    assert DEFAULT_APP_DIR == Path("tests/apps/working")
    assert DEFAULT_DATA_ROOT == Path("tests/data/working_1_no_mapping")
    assert "some_input" in INPUT_FILE_MAPPING


def test_process():
    FGProcess()


def test_process_raises():
    with pytest.raises(RuntimeError):
        FGProcess(data_dir="tests/data/working_1")



from io import StringIO
from logging import StreamHandler, WARNING, Logger
from contextlib import contextmanager
from typing import Callable, ContextManager
import fastgenomics.deprecated as deprecated
import pytest


@pytest.fixture
def catch_log_warnings() -> Callable[[Logger], ContextManager[StreamHandler]]:
    @contextmanager
    def catcher(logger: Logger):
        handler = StreamHandler(StringIO())
        handler.setLevel(WARNING)
        logger.addHandler(handler)
        yield handler

    return catcher


@pytest.fixture
def local(monkeypatch, app_dir, data_root):
    """patches the paths for local testing"""
    with pytest.deprecated_call():
        deprecated.set_paths(app_dir, data_root)


@pytest.fixture
def fg_env(monkeypatch, app_dir, data_root):
    """sets app_dir and data_root by env-variables"""
    monkeypatch.setenv("FG_APP_DIR", str(app_dir))
    monkeypatch.setenv("FG_DATA_ROOT", str(data_root))


@pytest.fixture
def clear_output(data_root):
    """clear everything except of .gitignore"""
    for name in ["output", "summary"]:
        sub_dir = data_root / name
        for entry in sub_dir.glob("*.*"):
            if entry.name != ".gitignore":
                entry.unlink()



from fastgenomics import deprecated
import pytest
import pathlib


def test_can_read_input_file(local):
    # can get path
    with pytest.deprecated_call():
        to_test = deprecated.get_input_path("some_input")

    # path exists
    assert to_test.exists()


def test_can_read_optional_input_file(local):
    with pytest.deprecated_call():
        to_test = deprecated.get_input_path("some_optional_input")
    assert to_test.exists()


def test_cannot_read_undefined_input(local):
    with pytest.deprecated_call():
        with pytest.raises(KeyError):
            deprecated.get_input_path("i_don't_exist")


def test_left_out_optional_is_none(app_dir, data_root_2):
    with pytest.deprecated_call():
        deprecated.set_paths(app_dir, data_root_2)
        assert None is deprecated.get_input_path("some_optional_input")


def test_can_write_summary(local, clear_output):
    with pytest.deprecated_call():
        sum_file = deprecated.get_summary_path()
    with sum_file.open("w", encoding="utf-8") as out:
        out.write("test")
    assert sum_file.exists()


def test_can_write_output(local, clear_output):
    with pytest.deprecated_call():
        out_path = deprecated.get_output_path("some_output")
    assert out_path.name == "some_output.csv"
    with out_path.open("w", encoding="utf-8") as out:
        out.write("test")
    assert out_path.exists()


def test_cannot_write_undefined_output(local):
    with pytest.deprecated_call():
        with pytest.raises(KeyError):
            deprecated.get_output_path("i_don't_exist")


# test things, imported from fastgenomics.common, are available
def test_import_from_common(app_dir, data_root):
    with pytest.deprecated_call():
        deprecated.set_paths(app_dir=app_dir, data_root=data_root)
        assert len(deprecated.get_parameters()) > 0
        deprecated.get_parameter("int_value")


def test_paths_are_initialized(local):
    with pytest.deprecated_call():
        deprecated.get_paths()


def test_custom_init_paths(app_dir, data_root):
    with pytest.deprecated_call():
        deprecated.set_paths(app_dir, data_root)
        deprecated.get_paths()


def test_paths_from_env(fg_env):
    with pytest.deprecated_call():
        deprecated.get_paths()


def test_cannot_init_nonexisting_paths():
    with pytest.deprecated_call():
        with pytest.raises(FileNotFoundError):
            deprecated.set_paths("i_don't_exist", "me_neither")


def test_get_app_manifest(local):
    with pytest.deprecated_call():
        deprecated.get_app_manifest()


def test_can_get_parameters(local):
    with pytest.deprecated_call():
        parameters = deprecated.get_parameters()
        assert len(parameters) > 0


def test_parameters(local):
    with pytest.deprecated_call():
        parameters = deprecated.get_parameters()

    assert "str_value" in parameters
    assert parameters["str_value"] == "hello from parameters.json"

    assert "int_value" in parameters
    assert parameters["int_value"] == 150

    assert "float_value" in parameters
    assert parameters["float_value"] == float(100)

    assert "bool_value" in parameters
    assert parameters["bool_value"] is True

    assert "list_value" in parameters
    assert parameters["list_value"] == [1, 2, 3]

    assert "dict_value" in parameters
    assert parameters["dict_value"] == {"foo": 42, "bar": "answer to everything"}

    assert "optional_int_value_concrete" in parameters
    assert parameters["optional_int_value_concrete"] == 4

    assert "optional_int_value_null" in parameters
    assert parameters["optional_int_value_null"] is None

    assert "enum_value" in parameters
    assert parameters["enum_value"] == "X"

    # moved to test_app_checker because logs are printed at a different time now
    # assert any(["Parameters" in x.message for x in caplog.records]), "Parameter logs are not set"


def test_can_get_specific_parameter(local):
    with pytest.deprecated_call():
        assert deprecated.get_parameter("int_value") == 150


def test_can_get_null_parameter(local):
    with pytest.deprecated_call():
        assert deprecated.get_parameter("optional_int_value_null") is None


def test_parameters_2(app_dir, data_root, data_root_2):
    with pytest.deprecated_call():
        deprecated.set_paths(app_dir, data_root)
        assert deprecated.get_parameter("int_value") == 150
        assert deprecated.get_parameter("str_value") == "hello from parameters.json"

        deprecated.set_paths(app_dir, data_root_2)
        assert (
            deprecated.get_parameter("str_value")
            == "hello from app 2's parameters.json"
        )


def test_input_file_mapping(app_dir, data_root, data_root_2):
    with pytest.deprecated_call():
        deprecated.set_paths(app_dir, data_root)
        assert deprecated.get_input_path("some_input").name == "expression_matrix.csv"

        deprecated.set_paths(app_dir, data_root_2)
        assert deprecated.get_input_path("some_input").name == "data"



import pytest
from pathlib import Path
from fastgenomics.process import FGProcess
from fastgenomics.testing import cleanoutput

HERE = Path(__file__).parent
DATA = HERE / "data"
APPS = HERE / "apps"


@pytest.fixture
def app_dir():
    return APPS / "working"


@pytest.fixture
def app_dir_validation_error():
    return APPS / "validation_error"


@pytest.fixture
def app_dir_missing_files():
    return APPS / "missing_files"


@pytest.fixture
def app_dir_non_existing():
    path = APPS / "non_existing"
    assert not path.exists()
    return path


@pytest.fixture
def app_dir_runtime_error():
    return APPS / "runtime_error"


@pytest.fixture
def data_root():
    return DATA / "working_1"


@pytest.fixture
def data_root_colnames():
    return DATA / "working_1_colnames"


@pytest.fixture
def data_root_2():
    return DATA / "working_2"


@pytest.fixture
def data_root_none():
    return DATA / "non_existing"


@pytest.fixture
def data_root_missing_files():
    return DATA / "missing_files"


@pytest.fixture
def data_root_no_mapping_file():
    return DATA / "no_mapping_file"


@pytest.fixture(params=["working_1", "working_3"])
def fgprocess(request, app_dir):
    fg = FGProcess(app_dir, DATA / request.param)
    cleanoutput(fg)
    return fg


@pytest.mark.anndata
@pytest.fixture
def adata(fgprocess):
    import fastgenomics.external.anndata as fgad

    return fgad.read(
        expr=fgprocess.input["some_input"],
        gene_meta=fgprocess.input["genes"],
        cell_meta=fgprocess.input["cells"],
    )


# data-specific fixtures for test_external
@pytest.fixture
def fgprocess_1(app_dir, data_root):
    fg = FGProcess(app_dir, data_root)
    cleanoutput(fg)
    return fg


# data-specific fixtures for test_external
@pytest.fixture
def fgprocess_1_colnames(app_dir, data_root_colnames):
    fg = FGProcess(app_dir, data_root_colnames)
    cleanoutput(fg)
    return fg


@pytest.fixture
def summary(fgprocess_1):
    fgprocess_1.summary.template = fgprocess_1.app.app_dir / "summary.md.j2"
    return fgprocess_1.summary


@pytest.mark.anndata
@pytest.fixture
def adata_1(fgprocess_1):
    import fastgenomics.external.anndata as fgad

    return fgad.read(
        expr=fgprocess_1.input["some_input"],
        gene_meta=fgprocess_1.input["genes"],
        cell_meta=fgprocess_1.input["cells"],
    )



def cleanoutput(fg):
    """Deletes all the output files.  Useful for cleanup before tests."""
    for file in fg.output.keys():
        path = fg.output[file].path
        if path.exists():
            path.unlink()



# coding: utf-8

from collections import OrderedDict
from pathlib import Path
import textwrap
from logging import getLogger
import os.path

from .defaults import SUMMARY_KEY

logger = getLogger("fastgenomics.common")


class Files(object):
    def __getitem__(self, key):
        if key not in self.files:
            raise KeyError(f'Key "{key}" not defined in manifest.json of the App')
        else:
            return self.files[key]

    def __contains__(self, key):
        return key in self.files

    def __repr__(self):
        output = "Files:\n"
        for file in self.files.values():
            output += textwrap.indent(file.__repr__(), "  ") + "\n"
        return output

    def keys(self):
        return self.files.keys()


class FilesInput(Files):
    def __init__(self, specs, root, mapping):
        self.files = OrderedDict()
        for name, spec in specs.items():
            path = None
            isoptional = "optional" in spec and spec["optional"]
            if name in mapping:
                path = Path(root) / mapping[name]
                if not path.exists():
                    raise FileNotFoundError(f"File {name}: not found under {path}")
            elif isoptional:
                path = None
            else:
                raise KeyError(
                    f'Non-optional file "{name}" missing from input_file_mapping.json'
                )
            self.files[name] = FileInput(
                name=name,
                type=spec["type"],
                usage=spec["usage"],
                path=path,
                optional=isoptional,
            )

        not_in_manifest = set(mapping) - set(specs)
        if not_in_manifest:
            logger.warning(
                f"Ignoring files defined in input_file_mapping.json: {not_in_manifest}"
            )

        logger.debug(f"Files found in the manifest: {list(mapping)}")
        logger.debug(f"Files found in the input_file_mapping.json: {list(specs)}")


class FilesOutput(Files):
    def __init__(self, specs, root, summary_path):
        self.files = OrderedDict()
        for name, spec in specs.items():
            self.files[name] = FileOutput(
                name=name,
                type=spec["type"],
                usage=spec["usage"],
                path=Path(root) / spec["file_name"],
            )
        self.files[SUMMARY_KEY] = FileOutput(
            name="summary", type="summary", usage="Summary file", path=summary_path
        )


class File(object):
    def __init__(self, name, type, usage, path):
        self.name = name
        self.type = type
        self.usage = usage
        self.path = path

    def __repr__(self):
        output = [f'"{arg.capitalize()}": {val}' for arg, val in vars(self).items()]
        return "\n".join(output)


class FileInput(File):
    def __init__(self, name, type, usage, path, optional=False):
        super().__init__(name, type, usage, path)
        self.optional = optional

    def __repr__(self):
        return "\n".join(
            [super().__repr__(), "IO type: Input", f"Optional: {self.optional}"]
        )

    def exists(self):
        return self.path is not None


class FileOutput(File):
    def __repr__(self):
        return "\n".join([super().__repr__(), "IO type: Output"])



import os
import json
from pathlib import Path


def get_mapping_from_env(var):
    mapping = os.environ.get(var, None)
    if mapping is None:
        return None
    else:
        return {
            key: os.path.expandvars(name) for key, name in json.loads(mapping).items()
        }


# set default paths
DEFAULT_APP_DIR = Path(os.path.expandvars(os.environ.get("FG_APP_DIR", "/app")))
DEFAULT_DATA_ROOT = Path(
    os.path.expandvars(os.environ.get("FG_DATA_ROOT", "/fastgenomics"))
)
INPUT_FILE_MAPPING = get_mapping_from_env("INPUT_FILE_MAPPING")


# summary key for the FGProcess.output
SUMMARY_KEY = "summary"



import jsonschema
import json
from pathlib import Path

from .defaults import DEFAULT_APP_DIR
from .parameters import FGParameters


class FGApp(object):
    # files that are checked for existence that are not otherwise
    # explicitly used by the FGApp
    _mandatory_files = ["LICENSE", "LICENSE-THIRD-PARTY", "manifest.json"]

    def __init__(self, app_dir=DEFAULT_APP_DIR):
        if isinstance(app_dir, str):
            app_dir = Path(app_dir)

        self.app_dir = app_dir
        self.check_files()

        self.manifest = self.get_manifest()
        self.application = self.manifest["application"]
        self.default_parameters = FGParameters(self.application["parameters"])
        self.app_type = self.application["type"]
        self.inputs = self.application["input"]
        self.outputs = self.application["output"]
        assert self.app_type in ["Calculation", "Visualization"]

    def get_manifest(self):
        manifest_file = self.app_dir / "manifest.json"
        manifest = json.loads(manifest_file.read_bytes())
        check_manifest(manifest)
        return manifest

    def check_files(self):
        if not self.app_dir.exists():
            raise FileNotFoundError(f"Could not find the App directory {self.app_dir}")

        not_found = []
        for f in self._mandatory_files:
            absolute_path = self.app_dir / f
            if not absolute_path.exists():
                not_found.append(f)

        if not_found:
            msg = [f'Could not find the following files in "{self.app_dir}":']
            msg += [f"- {file}" for file in not_found]
            raise FileNotFoundError("\n".join(msg))


def check_manifest(config: dict):
    """
    Asserts that the manifest (``manifest.json``) matches our JSON-Schema.
    If not a :py:exc:`jsonschema.ValidationError` will be raised.
    """

    schema_file = Path(__file__).parent / "schemes" / "manifest_schema.json"
    schema = json.loads(schema_file.read_text())
    jsonschema.validate(config, schema)

    input_types = config["application"]["input"]
    output_types = config["application"]["output"]

    def err_msg(x, y):
        return "'{}'-type not supported for {}-operations.".format(x, y)

    if input_types is not None:
        for name, properties in input_types.items():
            if properties["type"] == "output_only":
                raise RuntimeError(err_msg(properties["type"], "input"))

    if output_types is not None:
        for name, properties in output_types.items():
            if properties["type"] == "dataset_manifest":
                raise RuntimeError(err_msg(properties["type"], "output"))



import fastgenomics.external.anndata as fgad
import pytest
import pandas as pd
import numpy as np


@pytest.mark.anndata
def test_read(fgprocess_1):
    adata = fgad.read(
        expr=fgprocess_1.input["some_input"],
        gene_meta=fgprocess_1.input["genes"],
        cell_meta=fgprocess_1.input["cells"],
    )

    assert adata.obs.index.equals(pd.Index(["c1", "c3", "c4", "c2"]))
    assert adata.var.index.equals(
        pd.Index(["1", "3", "5", "10", "100", "unknown", "unknown", "ambiguous"])
    )
    assert adata.X[1, 0] == 4
    assert adata.shape == (4, 8)
    assert adata.var.shape[1] == 3
    assert adata.obs.shape[1] == 3
    assert "some_column" in adata.var
    assert "some_column" in adata.obs
    assert "batch_id" in adata.obs


@pytest.mark.anndata
def test_read_colnames(fgprocess_1_colnames):
    adata = fgad.read(
        expr=fgprocess_1_colnames.input["some_input"],
        gene_meta=fgprocess_1_colnames.input["genes"],
        cell_meta=fgprocess_1_colnames.input["cells"],
        gene_id="gene_id_1",
        cell_id="cell_id_1",
        expression_id="expression_id_1",
    )

    assert adata.obs.index.equals(pd.Index(["c1", "c3", "c4", "c2"]))
    assert adata.var.index.equals(
        pd.Index(["1", "3", "5", "10", "100", "unknown", "unknown", "ambiguous"])
    )
    assert adata.X[1, 0] == 4
    assert adata.shape == (4, 8)
    assert adata.var.shape[1] == 3
    assert adata.obs.shape[1] == 3
    assert "some_column" in adata.var
    assert "some_column" in adata.obs
    assert "batch_id" in adata.obs


@pytest.mark.anndata
def test_read_partial(fgprocess):
    fgad.read(expr=fgprocess.input["some_input"])
    fgad.read(expr=fgprocess.input["some_input"], gene_meta=fgprocess.input["genes"])
    fgad.read(expr=fgprocess.input["some_input"], cell_meta=fgprocess.input["cells"])
    fgad.read(
        expr=fgprocess.input["some_input"],
        cell_meta=fgprocess.input["non_existing_optional_input"],
        gene_meta=fgprocess.input["non_existing_optional_input"],
    )


@pytest.mark.anndata
def test_read_type_throws(fgprocess):
    with pytest.raises(TypeError):
        fgprocess.input["some_input"].type = "wrong_type"
        fgad.read(
            expr=fgprocess.input["some_input"],
            gene_meta=fgprocess.input["genes"],
            cell_meta=fgprocess.input["cells"],
        )


@pytest.mark.anndata
def test_write(fgprocess, adata):
    for f in ["count_output", "genes_output", "cells_output"]:
        assert not fgprocess.output[f].path.exists()

    fgad.write(
        adata,
        expr=fgprocess.output["count_output"],
        gene_meta=fgprocess.output["genes_output"],
        cell_meta=fgprocess.output["cells_output"],
    )

    for f in ["count_output", "genes_output", "cells_output"]:
        assert fgprocess.output[f].path.exists()


@pytest.mark.anndata
def test_read_write(fgprocess_1, adata_1):
    fgad.write(
        adata_1,
        expr=fgprocess_1.output["count_output"],
        gene_meta=fgprocess_1.output["genes_output"],
        cell_meta=fgprocess_1.output["cells_output"],
    )

    expr_out = pd.read_csv(
        fgprocess_1.output["count_output"].path,
        dtype={
            "entrez_id": "category",
            "cell_id": "category",
            "expression": np.float64,
        },
    )
    expr_out = expr_out.sort_values(list(expr_out.columns)).reset_index(drop=True)

    genes_out = pd.read_csv(
        fgprocess_1.output["genes_output"].path, dtype={"entrez_id": "category"}
    )
    genes_out.set_index("entrez_id", inplace=True, drop=False)

    cells_out = pd.read_csv(
        fgprocess_1.output["cells_output"].path, dtype={"cell_id": "category"}
    )
    cells_out.set_index("cell_id", inplace=True, drop=False)

    expr_in = pd.DataFrame(
        dict(
            cell_id=pd.Categorical(["c1", "c1", "c2", "c3", "c3"]),
            entrez_id=pd.Categorical(["1", "3", "5", "1", "100"]),
            expression=[1.0, 2.0, 3.0, 4.0, 1.0],
        )
    )
    expr_in = expr_in.sort_values(list(expr_out.columns)).reset_index(drop=True)

    assert expr_out.equals(expr_in)
    assert genes_out.equals(adata_1.var)
    assert cells_out.equals(adata_1.obs)


@pytest.mark.anndata
def test_preserve_column_order(fgprocess):
    adata = fgad.read(
        expr=fgprocess.input["some_input"],
        gene_meta=fgprocess.input["genes"],
        cell_meta=fgprocess.input["cells"],
    )

    fgad.write(
        adata,
        expr=fgprocess.output["count_output"],
        gene_meta=fgprocess.output["genes_output"],
        cell_meta=fgprocess.output["cells_output"],
    )

    assert (
        fgprocess.input["genes"].path.read_text()
        == fgprocess.output["genes_output"].path.read_text()
    )

    assert (
        fgprocess.input["cells"].path.read_text()
        == fgprocess.output["cells_output"].path.read_text()
    )



import jinja2
from pathlib import Path


class FGSummary(object):
    def __init__(self, output, params):
        if output.type != "summary":
            raise TypeError(f'Expected a summary file but got "{output.type}"')
        self.output = output.path
        self.params = params
        self.footnote = self.summary_footnote()
        self.template = None

    def summary_footnote(self):
        footnote = "\n### Parameters\n"
        for name, param in self.params.parameter.items():
            if param.type == str:
                value = f'"{param.value}"'
            else:
                value = param.value
            footnote += f"* __{name}__ = `{value}`  _({param.description})_\n"
        return footnote

    def render_template(self, temp_file, **kwargs):
        summary = temp_file.read_text() + self.footnote
        temp = jinja2.Template(summary, undefined=jinja2.StrictUndefined)
        return temp.render(kwargs)

    def write(self, **kwargs):
        if self.template is None:
            raise AttributeError(
                "Please specify the template before calling `write` by setting the `template` parameter."
            )

        template = Path(self.template)
        if not template.exists():
            raise FileNotFoundError(
                f'Could not find the summary template under "{template}".  You can change the template location by modifying the `template` argument.'
            )
        summary = self.render_template(template, **kwargs)
        self.output.write_text(summary)



from deprecated import deprecated
from .process import FGProcess

_PROCESS = None


def get_process():
    if _PROCESS is None:
        raise NameError(
            f"call set_paths(app_dir, data_root) before accessing the global process."
        )
    else:
        return _PROCESS


@deprecated(reason="Use FGProcess instead")
def set_paths(app_dir, data_root):
    global _PROCESS
    _PROCESS = FGProcess(app_dir, data_root)
    _PROCESS.data.paths["app"] = app_dir


@deprecated(reason="Use FGProcess.parameters instead")
def get_parameter(key):
    return get_process().parameters[key]


@deprecated(reason="Use FGProcess.input instead")
def get_input_path(filename):
    return get_process().input[filename].path


@deprecated(reason="Use FGProcess.output instead")
def get_output_path(filename):
    return get_process().output[filename].path


@deprecated(reason="Use FGProcess.summary or FGProcess.output['summary'] instead")
def get_summary_path():
    return get_process().output["summary"].path


@deprecated(reason="Use FGProcess.data.paths instead")
def get_paths():
    return {**get_process().data.paths, "app": get_process().app.app_dir}


@deprecated(reason="Use FGProcess.app.manifest instead")
def get_app_manifest():
    return get_process().app.manifest


@deprecated(reason="Use FGProcess.parameters instead")
def get_parameters():
    return {
        name: param.value for name, param in get_process().parameters.parameter.items()
    }



from fastgenomics.process import FGProcess
import pytest


def test_io_throws_on_non_existing_files(app_dir, data_root_none):
    with pytest.raises(FileNotFoundError):
        FGProcess(app_dir, data_root_none)


def test_undefined_for_optional_files(app_dir, data_root):
    fg = FGProcess(app_dir, data_root)
    optional = fg.input["non_existing_optional_input"]
    assert optional.path is None
    assert not optional.exists()



from .app import FGApp
from .data import FGData
from .io import FilesInput, FilesOutput
from .summary import FGSummary
from .defaults import DEFAULT_APP_DIR, DEFAULT_DATA_ROOT, SUMMARY_KEY

from logging import getLogger


class FGProcess(object):
    def __init__(self, app_dir=DEFAULT_APP_DIR, data_dir=DEFAULT_DATA_ROOT):
        self.data = FGData(data_dir)
        self.app = FGApp(app_dir)

        self.parameters = self.app.default_parameters.copy()
        self.parameters.update(self.data.parameters)

        self.logger = getLogger(f"fastgenomics.process")

        # log the updated parameter values
        info = "\n".join(f"{k}:{v.value}" for k, v in self.parameters.parameter.items())
        self.logger.info(f"Parameters: \n{info}")

        self.input = FilesInput(
            specs=self.app.inputs,
            root=self.data.paths["data"],
            mapping=self.data.input_file_mapping,
        )
        self.output = FilesOutput(
            specs=self.app.outputs,
            root=self.data.paths["output"],
            summary_path=self.data.paths[SUMMARY_KEY] / "summary.md",
        )

        self.summary = FGSummary(
            output=self.output[SUMMARY_KEY], params=self.parameters
        )



from pathlib import Path
from logging import getLogger
import json

from .defaults import DEFAULT_DATA_ROOT, INPUT_FILE_MAPPING

logger = getLogger(f"fastgenomics.data")

DATA_SUBDIRS = ["data", "config", "output", "summary"]
MANDATORY_SUBDIRS = ["data", "config"]


class FGData(object):
    """This class stores the paths to data structured according to the
fastgenomics specification.  It also loads the input file mappings and
checks if the files exist.

    """

    _subdirs = DATA_SUBDIRS
    _mandatory_files = MANDATORY_SUBDIRS

    def __init__(
        self, data_root=DEFAULT_DATA_ROOT, env_input_file_mapping=INPUT_FILE_MAPPING
    ):
        if isinstance(data_root, str):
            data_root = Path(data_root)
        self.root = data_root
        self.env_input_file_mapping = env_input_file_mapping
        self.check_files()

        self.paths = self.get_paths()
        self.input_file_mapping = self.get_input_file_mapping()
        self.parameters = self.get_parameters()

    def get_input_file_mapping(self):
        mapping_file = self.paths["config"] / "input_file_mapping.json"

        if self.env_input_file_mapping is not None and mapping_file.exists():
            raise RuntimeError(
                f'Environmental variable INPUT_FILE_MAPPING was defined and a mapping file "{mapping_file}" was found.  Please use just one of the abouve to define the input file mapping.'
            )
        elif self.env_input_file_mapping is not None:
            logger.debug(
                'Loading the file mapping from an environmental variable "INPUT_FILE_MAPPING".'
            )
            mapping = self.env_input_file_mapping
        elif mapping_file.exists():
            logger.debug(
                f'Loading the input file mapping from a file "{mapping_file}".  The current version of FASTGenomics runtime uses an environmental variable INPUT_FILE_MAPPING to pass the information on file mapping and the file "input_file_mapping.json" should only be used for testing.'
            )
            mapping = json.loads(mapping_file.read_text())
        else:
            raise RuntimeError(
                f'Environmental variable INPUT_FILE_MAPPING was undefined and a mapping file was not found under "{mapping_file}".  Please use one or the other to define the file mapping.'
            )

        logger.debug(f'File mapping is: "{mapping}"')
        return mapping

    def get_paths(self):
        return {dir: self.root / dir for dir in self._subdirs}

    def get_parameters(self):
        params_file = self.paths["config"] / "parameters.json"
        if params_file.exists():
            return json.loads(params_file.read_text())
        else:
            return {}

    def check_files(self):
        if not self.root.exists():
            raise FileNotFoundError(
                f"Could not find the data directory under {self.root}"
            )

        not_found = []
        for f in self._mandatory_files:
            absolute_path = self.root / f
            if not absolute_path.exists():
                not_found.append(f)

        if not_found:
            msg = [
                f'Could not find the following files or directories in "{self.root}":'
            ]
            msg += [f"- {file}" for file in not_found]
            raise FileNotFoundError("\n".join(msg))



try:
    import pandas as pd
    import numpy as np
    import scipy.sparse as sp
    from anndata import AnnData
except ImportError as e:
    msg = f"Could not import some of the necessary modules ({e.name}).  Please make sure to install anndata (https://github.com/theislab/anndata) with all its dependencies correctly (e.g. pandas, numpy, scipy)."
    raise ImportError(msg, name=e.name, path=e.path)
    raise e


EXPRESSION_MATRIX_DTYPE = {
    "entrez_id": "category",
    "cell_id": "category",
    "expression": np.float64,
}


def check_file_type(file, content_type):
    """Returns a path if the file is of requested FASTGenomics type,
otherwise throws a TypeError."""
    if file.type != content_type:
        raise TypeError(
            f'File "{file.name}" is of type "{file.type}" but expected "{content_type}".'
        )


def read(
    expr,
    cell_meta=None,
    gene_meta=None,
    cell_id="cell_id",
    gene_id="entrez_id",
    expression_id="expression",
):
    """Reads an anndata object by composing three files together:
`expression_matrix`, `cell_metadata` and `gene_metadata`."""
    check_file_type(expr, content_type="expression_matrix")
    expr = pd.read_csv(
        expr.path,
        dtype={cell_id: "category", gene_id: "category", expression_id: np.float32},
    )

    obs = read_cell_metadata(cell_meta, expr, cell_id)
    var = read_gene_metadata(gene_meta, expr, gene_id)

    counts = read_sparse_matrix(expr, obs, var, cell_id, gene_id, expression_id)

    adata = AnnData(counts, obs=obs, var=var, dtype="float32")
    return adata


def read_cell_metadata(cell_meta, expr, cell_id):
    expr_cell_id = expr[cell_id].unique()

    if cell_meta is None or not cell_meta.exists():
        df = pd.DataFrame(data={cell_id: expr_cell_id})
    else:
        check_file_type(cell_meta, content_type="cell_metadata")
        df = pd.read_csv(cell_meta.path, dtype={cell_id: str})

    df[cell_id] = df[cell_id].astype("category")
    df.set_index(cell_id, inplace=True, drop=False)

    missing_ids = set(expr_cell_id) - set(df.index)
    if missing_ids:
        raise Exception(
            f'Some {cell_id}\'s were present in the expression matrix but not in "{cell_meta.path}": {missing_ids}.'
        )
    return df


def read_gene_metadata(gene_meta, expr, gene_id):
    expr_entrez_id = expr[gene_id].unique()

    if gene_meta is None or not gene_meta.exists():
        df = pd.DataFrame(data={gene_id: expr_entrez_id})
    else:
        check_file_type(gene_meta, content_type="gene_metadata")
        df = pd.read_csv(gene_meta.path, dtype={gene_id: str})

    df[gene_id] = df[gene_id].astype("category")
    df.set_index(gene_id, inplace=True, drop=False)

    missing_ids = set(expr_entrez_id) - set(df.index)
    if missing_ids:
        raise Exception(
            f'Some {gene_id}\'s were present in the expression matrix but not in "{gene_meta.path}": {missing_ids}.'
        )

    return df


def read_sparse_matrix(expr, obs, var, cell_id, gene_id, expression_id):
    cell_idx = pd.DataFrame(
        {cell_id: obs.index, "cell_idx": np.arange(obs.shape[0])}
    ).set_index(cell_id)
    gene_idx = pd.DataFrame(
        {gene_id: var.index, "gene_idx": np.arange(var.shape[0])}
    ).set_index(gene_id)
    expr = expr.merge(cell_idx, on=cell_id, copy=False)
    expr = expr.merge(gene_idx, on=gene_id, copy=False)

    counts = sp.coo_matrix(
        (expr[expression_id], (expr.cell_idx, expr.gene_idx)),
        shape=(obs.shape[0], var.shape[0]),
    ).tocsr()

    return counts


# Writing
def write_exprs_csv(adata, csv_file):
    mat = adata.X.tocoo()
    df = pd.DataFrame.from_dict(
        dict(
            cell_id=adata.obs_names[mat.row],
            entrez_id=adata.var_names[mat.col],
            expression=mat.data,
        )
    )
    df.to_csv(csv_file, index=False)


def write(adata, expr=None, cell_meta=None, gene_meta=None):
    if expr is not None:
        check_file_type(expr, content_type="expression_matrix")
        write_exprs_csv(adata, expr.path)

    if cell_meta is not None:
        check_file_type(cell_meta, content_type="cell_metadata")
        adata.obs.to_csv(cell_meta.path, index=False)

    if gene_meta is not None:
        check_file_type(gene_meta, content_type="gene_metadata")
        adata.var.to_csv(gene_meta.path, index=False)
